import unittest

from pyramid.config import Configurator

from .. import DualCookieCSRFStoragePolicy
from pyramid.interfaces import ICSRFStoragePolicy


class DummyRequest(object):
    registry = None
    session = None
    response_callbacks = None
    scheme = None

    def __init__(self, registry=None, session=None, scheme='http'):
        self.registry = registry
        self.session = session
        self.cookies = {}
        self.scheme = scheme
        self.response_callbacks = []

    def add_response_callback(self, callback):
        self.response_callbacks.append(callback)

    def response_callback(self, request, response):
        while len(self.response_callbacks):
            callback = self.response_callbacks.pop()
            callback(request, response)


class MockResponse(object):
    def __init__(self):
        self.headerlist = []


class TestCookieCSRFStoragePolicy_InsecureHttp(unittest.TestCase):
    """tests on the insecure channel"""

    def _makeOne(self, **kw):
        return DualCookieCSRFStoragePolicy(**kw)

    def test_register_cookie_csrf_policy(self):
        config = Configurator()
        config.set_csrf_storage_policy(self._makeOne())
        config.commit()
        policy = config.registry.queryUtility(ICSRFStoragePolicy)
        self.assertTrue(isinstance(policy, DualCookieCSRFStoragePolicy))

    def test_get_cookie_csrf_with_no_existing_cookie_sets_cookies(self):
        response = MockResponse()
        request = DummyRequest()

        policy = self._makeOne()
        token = policy.get_csrf_token(request)
        request.response_callback(request, response)
        self.assertEqual(
            response.headerlist,
            [('Set-Cookie', 'csrf_http={}; Path=/; SameSite=Lax'.format(token))]
        )

    def test_get_cookie_csrf_nondefault_samesite(self):
        response = MockResponse()
        request = DummyRequest()

        policy = self._makeOne(samesite=None)
        token = policy.get_csrf_token(request)
        request.response_callback(request, response)
        self.assertEqual(
            response.headerlist,
            [('Set-Cookie', 'csrf_http={}; Path=/'.format(token))],
        )
        
    def test_existing_cookie_csrf_does_not_set_cookie(self):
        request = DummyRequest()
        request.cookies = {'csrf_http': 'e6f325fee5974f3da4315a8ccf4513d2'}

        policy = self._makeOne()
        token = policy.get_csrf_token(request)

        self.assertEqual(
            token,
            'e6f325fee5974f3da4315a8ccf4513d2'
        )
        self.assertEqual(len(request.response_callbacks), 0)

    def test_new_cookie_csrf_with_existing_cookie_sets_cookies(self):
        request = DummyRequest()
        request.cookies = {'csrf_http': 'e6f325fee5974f3da4315a8ccf4513d2'}

        policy = self._makeOne()
        token = policy.new_csrf_token(request)

        response = MockResponse()
        request.response_callback(request, response)
        self.assertEqual(
            response.headerlist,
            [('Set-Cookie', 'csrf_http={}; Path=/; SameSite=Lax'.format(token))]
        )

    def test_get_csrf_token_returns_the_new_token(self):
        request = DummyRequest()
        request.cookies = {'csrf_http': 'foo'}

        policy = self._makeOne()
        self.assertEqual(policy.get_csrf_token(request), 'foo')

        token = policy.new_csrf_token(request)
        self.assertNotEqual(token, 'foo')
        self.assertEqual(token, policy.get_csrf_token(request))

    def test_check_csrf_token(self):
        request = DummyRequest()

        policy = self._makeOne()
        self.assertFalse(policy.check_csrf_token(request, 'foo'))

        request.cookies = {'csrf_http': 'foo'}
        self.assertTrue(policy.check_csrf_token(request, 'foo'))
        self.assertFalse(policy.check_csrf_token(request, 'bar'))


class TestCookieCSRFStoragePolicy_SecureHTTPS(unittest.TestCase):
    """tests on the SECURE HTTPS"""

    def _makeOne(self, **kw):
        return DualCookieCSRFStoragePolicy(**kw)

    def test_register_cookie_csrf_policy(self):
        config = Configurator()
        config.set_csrf_storage_policy(self._makeOne())
        config.commit()
        policy = config.registry.queryUtility(ICSRFStoragePolicy)
        self.assertTrue(isinstance(policy, DualCookieCSRFStoragePolicy))

    def test_get_cookie_csrf_with_no_existing_cookie_sets_cookies(self):
        response = MockResponse()
        request = DummyRequest(scheme='https')

        policy = self._makeOne()
        token_secure = policy.get_csrf_token(request)
        token_http = policy.get_csrf_token_scheme(request, 'http')
        request.response_callback(request, response)

        self.assertEqual(
            response.headerlist,
            [('Set-Cookie', 'csrf_https={}; Path=/; secure; SameSite=Lax'.format(token_secure)),
             ('Set-Cookie', 'csrf_http={}; Path=/; SameSite=Lax'.format(token_http))
             ]
        )

    def test_get_cookie_csrf_nondefault_samesite(self):
        response = MockResponse()
        request = DummyRequest(scheme='https')

        policy = self._makeOne(samesite=None)
        token_secure = policy.get_csrf_token(request)
        token_http = policy.get_csrf_token_scheme(request, 'http')
        request.response_callback(request, response)
        self.assertEqual(
            response.headerlist,
            [('Set-Cookie', 'csrf_https={}; Path=/; secure'.format(token_secure)),
             ('Set-Cookie', 'csrf_http={}; Path=/'.format(token_http))
             ]
        )

    def test_existing_cookie_csrf_does_not_set_cookie(self):
        request = DummyRequest(scheme='https')
        request.cookies = {'csrf_https': 'e6f325fee5974f3da4315a8ccf4513d2'}

        policy = self._makeOne()
        token_secure = policy.get_csrf_token(request)

        self.assertEqual(
            token_secure,
            'e6f325fee5974f3da4315a8ccf4513d2'
        )
        self.assertEqual(len(request.response_callbacks), 0)

    def test_new_cookie_csrf_with_existing_cookie_sets_cookies(self):
        request = DummyRequest(scheme='https')
        request.cookies = {'csrf_https': 'e6f325fee5974f3da4315a8ccf4513d2'}

        policy = self._makeOne()
        token_secure = policy.new_csrf_token(request)
        token_http = policy.get_csrf_token_scheme(request, 'http')

        response = MockResponse()
        request.response_callback(request, response)
        self.assertEqual(
            response.headerlist,
            [('Set-Cookie', 'csrf_https={}; Path=/; secure; SameSite=Lax'.format(token_secure)),
             ('Set-Cookie', 'csrf_http={}; Path=/; SameSite=Lax'.format(token_http))
             ]
        )

    def test_get_csrf_token_returns_the_new_token(self):
        request = DummyRequest(scheme='https')
        request.cookies = {'csrf_https': 'foo'}

        policy = self._makeOne()
        self.assertEqual(policy.get_csrf_token(request), 'foo')

        token_secure = policy.new_csrf_token(request)
        self.assertNotEqual(token_secure, 'foo')
        self.assertEqual(token_secure, policy.get_csrf_token(request))

    def test_check_csrf_token(self):
        request = DummyRequest(scheme='https')

        policy = self._makeOne()
        self.assertFalse(policy.check_csrf_token(request, 'foo'))

        request.cookies = {'csrf_https': 'foo'}
        self.assertTrue(policy.check_csrf_token(request, 'foo'))
        self.assertFalse(policy.check_csrf_token(request, 'bar'))



import uuid

from webob.cookies import CookieProfile
from zope.interface import implementer

from pyramid.compat import bytes_
from pyramid.compat import text_
from pyramid.interfaces import ICSRFStoragePolicy
from pyramid.util import strings_differ
from pyramid.util import SimpleSerializer


__VERSION__ = '0.0.6'


# ==============================================================================


@implementer(ICSRFStoragePolicy)
class DualCookieCSRFStoragePolicy(object):
    """
    A re-implementation of ``pyramid.csrf.CookieCSRFStoragePolicy``.
    This policy runs separate csrf tokens on http and https connections.

    If the connection is https:
        expiring the csrf will expire the http and https
        only the https csrf will be consulted

    If the connection is http:
        expiring the csrf will expire the http
        only the http csrf will be consulted
    
    last synced to pyramid: 1.10.4
    
    NOTE: this does not support a ``secure`` argument
    """
    _token_factory = staticmethod(lambda: text_(uuid.uuid4().hex))

    def __init__(
        self,
        cookie_name_secure='csrf_https',
        cookie_name_http='csrf_http',
        secure=None,
        httponly=False,
        domain=None,
        max_age=None,
        path='/',
        samesite='Lax',
    ):
        if secure is not None:
            raise ValueError("`DualCookieCSRFStoragePolicy` does not support "
                             "the `secure` argument.")
        serializer = SimpleSerializer()
        self.cookie_name_secure = cookie_name_secure
        self.cookie_name_http = cookie_name_http
        self.cookie_profile_secure = CookieProfile(
            cookie_name=cookie_name_secure,
            secure=True,
            max_age=max_age,
            httponly=httponly,
            path=path,
            domains=[domain],
            serializer=serializer,
            samesite=samesite,
        )
        self.cookie_profile_http = CookieProfile(
            cookie_name=cookie_name_http,
            secure=False,
            max_age=max_age,
            httponly=httponly,
            path=path,
            domains=[domain],
            serializer=serializer,
            samesite=samesite,
        )

    @property
    def cookie_profile(self):
        """
        Returns the active cookie profile for the request's scheme.
        """
        raise ValueError("`DualCookieCSRFStoragePolicy` supports access "
                         "via the attributes `.cookie_profile_secure` and "
                         "via the attributes `.cookie_profile_http`. The "
                         "attribute `.cookie_profile` is not supported."
                         )

    def new_csrf_token(self, request):
        """ Sets a new CSRF token into the request and returns it. """

        def _token_secure():
            token = self._token_factory()
            request.cookies[self.cookie_name_secure] = token

            def _set_cookie(request, response):
                self.cookie_profile_secure.set_cookies(
                    response,
                    token,
                )
            request.add_response_callback(_set_cookie)
            return token

        def _token_http():
            token = self._token_factory()
            request.cookies[self.cookie_name_http] = token

            def _set_cookie(request, response):
                self.cookie_profile_http.set_cookies(
                    response,
                    token,
                )
            request.add_response_callback(_set_cookie)
            return token

        # if we are on https, reset the http and https
        # but only consult the https for data
        if request.scheme == 'https':
            token_http = _token_http()  # noqa
            token_secure = _token_secure()
            return token_secure

        # otherwise, set only the http insecure cookie
        token_http = _token_http()
        return token_http

    def get_csrf_token(self, request):
        """
        Returns the currently active CSRF token by checking the cookies
        sent with the current request.
        """
        if request.scheme == 'https':
            # only consult the secure cookie
            bound_cookies = self.cookie_profile_secure.bind(request)
        else:
            # the only accessible cookie is http insecure
            bound_cookies = self.cookie_profile_http.bind(request)
        token = bound_cookies.get_value()
        if not token:
            token = self.new_csrf_token(request)
        return token

    def check_csrf_token(self, request, supplied_token):
        """ Returns ``True`` if the ``supplied_token`` is valid."""
        expected_token = self.get_csrf_token(request)
        return not strings_differ(
            bytes_(expected_token), bytes_(supplied_token))

    def get_csrf_token_scheme(self, request, scheme):
        """this is a utility for testing"""
        if scheme not in ('https', 'http'):
            raise ValueError("unknown scheme")
        if scheme == 'https':
            if request.scheme != 'https':
                return None
            bound_cookies = self.cookie_profile_secure.bind(request)
            token = bound_cookies.get_value()
            return token
        bound_cookies = self.cookie_profile_http.bind(request)
        token = bound_cookies.get_value()
        return token



##############################################################################
#
# Copyright (c) 2017 Jonathan Vanasco and Contributors
# Portions Copyright (c) 2008-2013 Agendaless Consulting and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the BSD-like license at
# http://www.repoze.org/LICENSE.txt.  A copy of the license should accompany
# this distribution.  THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL
# EXPRESS OR IMPLIED WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND
# FITNESS FOR A PARTICULAR PURPOSE
#
##############################################################################
import os

from setuptools import setup
from setuptools import find_packages

# store version in the init.py
import re
with open(
        os.path.join(
            os.path.dirname(__file__),
            'pyramid_csrf_multi_scheme', '__init__.py')) as v_file:
    VERSION = re.compile(
        r".*__VERSION__ = '(.*?)'",
        re.S).match(v_file.read()).group(1)

try:
    here = os.path.abspath(os.path.dirname(__file__))
    README = open(os.path.join(here, "README.md")).read()
    README = README.split("\n\n", 1)[0] + "\n"
except:
    README = ''

requires = [
    "pyramid>=1.10.4",
]

setup(
    name="pyramid_csrf_multi_scheme",
    version=VERSION,
    description="provides for creating independent csrf tokens for the http and https schemes",
    long_description=README,
    classifiers=[
        "Intended Audience :: Developers",
        "Framework :: Pyramid",
        "Programming Language :: Python",
        "License :: Repoze Public License",
    ],
    keywords="web pyramid csrf",
    packages=['pyramid_csrf_multi_scheme',
              'pyramid_csrf_multi_scheme.tests',
              ],
    author="Jonathan Vanasco",
    author_email="jonathan@findmeon.com",
    url="https://github.com/jvanasco/pyramid_csrf_multi_scheme",
    license="BSD-derived (http://www.repoze.org/LICENSE.txt)",
    include_package_data=True,
    zip_safe=False,
    tests_require = requires,
    install_requires = requires,
    test_suite="pyramid_csrf_multi_scheme.tests",
)
